@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2007-2019 Sergey Poznyakoff
+   Copyright (C) 2007-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -479,13 +479,13 @@ json_object_set_bool (struct json_value *obj, char const *name, int val)
 }
 
 static struct json_value *
-json_error_reply_create (const char *msg)
+json_error_reply_create (int http_code, const char *msg)
 {
   struct json_value *val;
   
   val = json_reply_create ();
-  json_object_set_string (val, "status", "ER");
-  json_object_set_string (val, "error_message", "%s", msg);
+  json_object_set_string (val, "status", http_code < 300 ? "OK" : "ER");
+  json_object_set_string (val, "message", "%s", msg);
   return val;
 }
 
@@ -607,11 +607,11 @@ ctlio_reply (struct ctlio *io, int code, const char *fmt, ...)
       grecs_vasprintf (&str, &len, fmt, ap);
       va_end (ap);
 
-      io->output.reply = json_error_reply_create (str);
+      io->output.reply = json_error_reply_create (code, str);
       grecs_free (str);
     }
   else
-    io->output.reply = json_error_reply_create (http_text (code));
+    io->output.reply = json_error_reply_create (code, http_text (code));
 }
 
 static void
@@ -918,14 +918,19 @@ static void res_instance (struct ctlio *, enum http_method, char const *,
 			  struct json_value *);
 static void res_programs (struct ctlio *, enum http_method, char const *,
 			  struct json_value *);
+static void res_alive (struct ctlio *, enum http_method, char const *,
+		       struct json_value *);
+static void res_conf (struct ctlio *, enum http_method, char const *,
+		      struct json_value *);
+
+#if PIES_SYSVINIT_ENABLED
 static void res_runlevel (struct ctlio *, enum http_method, char const *,
 			  struct json_value *);
 static void res_environ  (struct ctlio *, enum http_method, char const *,
 			  struct json_value *);
-static void res_conf (struct ctlio *, enum http_method, char const *,
-		      struct json_value *);
-
 static int pred_sysvinit (void);
+#endif
+
 
 struct ctlio_resource
 {
@@ -943,8 +948,12 @@ static struct ctlio_resource restab[] = {
   { S(/conf),     CTL_ADMIN_STATE, NULL, res_conf },
   { S(/programs), CTL_ADMIN_STATE|CTL_USER_STATE, NULL,
     res_programs },
+  { S(/alive),    CTL_ADMIN_STATE|CTL_USER_STATE, NULL,
+    res_alive },
+#if PIES_SYSVINIT_ENABLED
   { S(/runlevel), CTL_ADMIN_STATE, pred_sysvinit, res_runlevel },
   { S(/environ),  CTL_ADMIN_STATE, pred_sysvinit, res_environ },
+#endif
   { NULL }
 #undef S
 };
@@ -1269,7 +1278,7 @@ ctl_accept (int socket, void *data)
   io = ctlio_create ();
   io->addr = addr;
   io->addrlen = addrlen;
-  register_socket (fd, NULL, ctlwr, NULL, io);
+  register_socket (fd, NULL, ctlwr, NULL, io, NULL);
 
   return 0;
 }
@@ -1296,7 +1305,7 @@ ctl_open (void)
       return -1;
     }
 
-  register_socket (fd, ctl_accept, NULL, NULL, NULL);
+  register_socket (fd, ctl_accept, NULL, NULL, NULL, NULL);
   
   return 0;
 }
@@ -1424,9 +1433,6 @@ auth_prog (struct prog *prog, struct ctlio *io)
     return CTL_ADMIN_STATE|CTL_USER_STATE;
   switch (prog->type)
     {
-    case TYPE_REDIRECTOR:
-      prog = prog->v.r.master;
-      /* FALL THROUGH */
     case TYPE_COMPONENT:
       if (prog->v.p.comp->adm_acl
 	  && check_acl (prog->v.p.comp->adm_acl,
@@ -1445,7 +1451,6 @@ auth_prog (struct prog *prog, struct ctlio *io)
 
 static char * const pies_type_str[] = {
   [TYPE_COMPONENT]   = "component", 
-  [TYPE_REDIRECTOR]  = "redirector", 
   [TYPE_COMMAND]     = "command"
 };
 
@@ -1894,15 +1899,11 @@ prog_serialize (struct json_value *ret, struct prog *prog)
 				prog->v.p.timestamp + SLEEPTIME);
 
       v = json_new_array ();
-      for (i = 0; i < prog->v.p.comp->argc; i++)
-	json_array_append (v, json_new_string (prog->v.p.comp->argv[i]));
+      for (i = 0; i < prog->v.p.argc; i++)
+	json_array_append (v, json_new_string (prog->v.p.argv[i]));
       json_object_set (ret, "argv", v);
       break;
 	  
-    case TYPE_REDIRECTOR:
-      json_object_set_number (ret, "PID", prog->pid);
-      break;
-	  
     case TYPE_COMMAND:
       json_object_set_string (ret, "command", "%s", prog->v.c.command);
     }
@@ -2116,7 +2117,59 @@ res_programs (struct ctlio *io, enum http_method meth,
   else
     res_programs_select (io, meth, uri, json);
 }
+
+static void
+res_alive (struct ctlio *io, enum http_method meth, char const *uri,
+	   struct json_value *json)
+{
+  struct prog *prog;
+	
+  if (!uri)
+    {
+      ctlio_reply (io, 403, NULL);
+      return;
+    }
+
+  prog = progman_locate (uri + 1);
+
+  if (prog && IS_COMPONENT (prog))
+    {
+      switch (prog->v.p.status)
+	{
+	case status_running:
+	case status_listener:
+	  /* FIXME: Presumably these three fall into that category too:
+	     status_stopped   -- Component is stopped administratively
+	     status_stopping  -- Component is being stopped
+	     status_finished  -- A "once" or "startup" component has finished
+	  */
+	  ctlio_reply (io, 200, NULL);
+	  break;
+
+	case status_sleeping:
+	  {
+	    time_t t;
+	    char buf[80];
+
+	    t = time (NULL) - prog->v.p.timestamp;
+	    if (t <= SLEEPTIME)
+	      {
+		snprintf (buf, sizeof buf, "%lu", SLEEPTIME - t);
+		output_set_header (&io->output, "Retry-After", buf);
+	      }
+	  }
+	  /* fall through */
+	  
+	default:	  
+	  ctlio_reply (io, 503, NULL);
+	  break;
+	}
+    }
+  else
+    ctlio_reply (io, 404, NULL);
+}
 
+#if PIES_SYSVINIT_ENABLED
 static int
 pred_sysvinit (void)
 {
@@ -2163,6 +2216,89 @@ res_runlevel (struct ctlio *io, enum http_method meth,
   else
     ctlio_reply (io, 405, NULL);
 }
+
+/* GET /environ          -     List entire environment
+ *   ["RUNLEVEL=3", "CONSOLE=/dev/tty", ...]  
+ * GET /environ/NAME     -     Get value of variable NAME
+ *   { "status":"OK", "value":"..." }
+ *   { "status":"ER", "error_message":"..." }
+ * DELETE /environ/NAME  -     Unset variable
+ *   { "status":"OK" }
+ *   { "status":"ER", "error_message":"..." }
+ * PUT /environ/NAME=VALUE -   Set variable 
+ *   { "status":"OK" }
+ *   { "status":"ER", "error_message":"..." }
+ */
+static void
+env_reply (struct ctlio *io, int ok, int rc)
+{
+  switch (rc)
+    {
+    case 0:
+      io->code = ok;
+      io->output.reply = json_reply_create ();
+      json_object_set_string (io->output.reply, "status", "OK");
+      break;
+
+    case 1:
+      ctlio_reply (io, 403, NULL);
+      break;
+      
+    case -1:
+      ctlio_reply (io, 404, NULL);
+    }
+}
+
+static void
+res_environ (struct ctlio *io, enum http_method meth,
+	     char const *uri, struct json_value *json)
+{
+  if (meth == METH_GET)
+    {
+      if (uri && uri[1])
+	{
+	  char *value;
+	  
+	  if (sysvinit_envlocate (uri + 1, &value) == -1)
+	    ctlio_reply (io, 404, NULL);
+	  else
+	    {
+	      env_reply (io, 200, 0);
+	      json_object_set_string (io->output.reply, "value", "%s", value);
+	    }
+	}
+      else
+	{
+	  size_t i;
+      
+	  io->output.reply = json_new_array ();
+	  io->code = 200;
+	  for (i = 0; sysvinit_environ_hint[i]; i++)
+	    {
+	      json_array_append (io->output.reply,
+				 json_new_string (sysvinit_environ_hint[i]));
+	    }
+	}
+    }
+  else if (meth == METH_DELETE)
+    {
+      if (!(uri && uri[0]))
+	ctlio_reply (io, 400, NULL);
+      else
+	env_reply (io, 200, sysvinit_envupdate (uri + 1));
+    }
+  else if (meth == METH_PUT)
+    {
+      if (!(uri && uri[0]))
+	ctlio_reply (io, 400, NULL);
+      else
+	env_reply (io, 201, sysvinit_envupdate (uri + 1));
+    }
+  else
+    ctlio_reply (io, 405, NULL);
+    
+}
+#endif
 
 /* GET /conf/runtime     -     List configuration
  * PUT /conf/runtime     -     Reload configuration
@@ -2407,85 +2543,4 @@ res_conf (struct ctlio *io, enum http_method meth,
   else 
     ctlio_reply (io, 404, NULL);
 }
-
-/* GET /environ          -     List entire environment
- *   ["RUNLEVEL=3", "CONSOLE=/dev/tty", ...]  
- * GET /environ/NAME     -     Get value of variable NAME
- *   { "status":"OK", "value":"..." }
- *   { "status":"ER", "error_message":"..." }
- * DELETE /environ/NAME  -     Unset variable
- *   { "status":"OK" }
- *   { "status":"ER", "error_message":"..." }
- * PUT /environ/NAME=VALUE -   Set variable 
- *   { "status":"OK" }
- *   { "status":"ER", "error_message":"..." }
- */
-static void
-env_reply (struct ctlio *io, int ok, int rc)
-{
-  switch (rc)
-    {
-    case 0:
-      io->code = ok;
-      io->output.reply = json_reply_create ();
-      json_object_set_string (io->output.reply, "status", "OK");
-      break;
 
-    case 1:
-      ctlio_reply (io, 403, NULL);
-      break;
-      
-    case -1:
-      ctlio_reply (io, 404, NULL);
-    }
-}
-
-static void
-res_environ (struct ctlio *io, enum http_method meth,
-	     char const *uri, struct json_value *json)
-{
-  if (meth == METH_GET)
-    {
-      if (uri && uri[1])
-	{
-	  char *value;
-	  
-	  if (sysvinit_envlocate (uri + 1, &value) == -1)
-	    ctlio_reply (io, 404, NULL);
-	  else
-	    {
-	      env_reply (io, 200, 0);
-	      json_object_set_string (io->output.reply, "value", "%s", value);
-	    }
-	}
-      else
-	{
-	  size_t i;
-      
-	  io->output.reply = json_new_array ();
-	  io->code = 200;
-	  for (i = 0; sysvinit_environ_hint[i]; i++)
-	    {
-	      json_array_append (io->output.reply,
-				 json_new_string (sysvinit_environ_hint[i]));
-	    }
-	}
-    }
-  else if (meth == METH_DELETE)
-    {
-      if (!(uri && uri[0]))
-	ctlio_reply (io, 400, NULL);
-      else
-	env_reply (io, 200, sysvinit_envupdate (uri + 1));
-    }
-  else if (meth == METH_PUT)
-    {
-      if (!(uri && uri[0]))
-	ctlio_reply (io, 400, NULL);
-      else
-	env_reply (io, 201, sysvinit_envupdate (uri + 1));
-    }
-  else
-    ctlio_reply (io, 405, NULL);
-    
-}
