@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies.
-   Copyright (C) 2009-2019 Sergey Poznyakoff
+   Copyright (C) 2009-2020 Sergey Poznyakoff
 
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -96,7 +96,7 @@ inetd_conf_file (const char *file)
     }
 
   wsflags = WRDSF_NOVAR | WRDSF_NOCMD | WRDSF_WS | WRDSF_SQUEEZE_DELIMS;
-  while (getline (&buf, &size, fp) >= 0)
+  while (grecs_getline (&buf, &size, fp) >= 0)
     {
       char *p;      
       struct component *comp;
